freeze;

// generic type for diagonal quadratic forms
declare type DiagQuadFrm;
declare attributes DiagQuadFrm: coeffs, dimension, base_ring, disc;
declare verbose DQF_verbosity, 3;

import "helpers/varia.m" : ZZ, QQ;

// basic contructor
intrinsic DiagonalQuadraticForm( K::Rng, D::[RngElt] ) -> DiagQuadFrm
{
Given a sequence of coefficients D = [d1, d2,..., dn], constructs a diagonal form\
q = ( d1, d2, ..., dn ) ~ d1*x1^2 + ... dn*xn^2.
}
    require forall{ d : d in D | IsCoercible(K, d) }: "Coefficients must be coercible to the base field.";
    q := New(DiagQuadFrm);
    q`coeffs := [ K!d : d in D ];
    q`dimension := #D;
    q`base_ring := K;
    return q;
end intrinsic;

intrinsic DiagonalQuadraticForm( D::[RngElt] ) -> DiagQuadFrm
{ " } //"
    return DiagonalQuadraticForm( Universe(D), D );
end intrinsic;

intrinsic HyperbolicForm( K::Rng, n::RngIntElt ) -> DiagQuadFrm
{
Construct a hyperbolic form over K of dimension n\\ 
(i.e. an orthogonal sum of n/2 hyperbolic planes).
}
    require IsEven(n): "The second argument must be even.";
    return DiagonalQuadraticForm(K, [K!1 : j in [1..n/2]] cat [K!-1: j in [1..n/2]]);
end intrinsic;

// printing
intrinsic Print( q::DiagQuadFrm )
{Print q}
    printf "(";
    d := Dimension(q);
    for i in [1..d] do 
        if i gt 1 then printf ", "; end if;
        printf "%o", q`coeffs[i];
    end for;
    printf ")";
end intrinsic;

// basic conversions
intrinsic QuadraticFormPolynomial( q::DiagQuadFrm ) -> RngPolElt
{
Converts a diagonal quadratic (a1,...,ad) form to a polynomial 
a1*x1^2 + ... + ad*xd^2;
}
    d := Dimension(q);
    P := PolynomialRing(q`base_ring, d);
    return &+ [ q`coeffs[j]*P.j^2 : j in [1..d] ];
end intrinsic;

/*
intrinsic IsCoercible( P::RngMPol, q::DiagQuadFrm ) -> BoolElt, .
{
Return whether q is coercible into a polynomial ring P and the result if so
}
    p := QuadraticFormPolynomial(q);
    return IsCoercible(P, p);
end intrinsic;
*/

intrinsic QuadraticSpace( q::DiagQuadFrm ) -> ModTupRng
{
Convert a diagonal quadratic form to a quadratic space.
This function is equivalent to QuadraticSpace(QuadraticFormPolynomial(q))
}
    return QuadraticSpace(QuadraticFormPolynomial(q));
end intrinsic;

intrinsic QuadraticFormMatrix( q::DiagQuadFrm ) -> ModAlgElt
{
Convert a diagonal quadratic form into a diagonal matrix representing
this form. 
}
    return DiagonalMatrix( q`coeffs );
end intrinsic;

// diagonalization of a quadratic space
intrinsic Diagonalization( V::ModTupFld ) -> DiagQuadFrm
{
Given a quadratic space V, this function constructs a diagonal quadratic form isometric to V.
}
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";

    M := QuadraticFormMatrix(V);
    D := Diagonalization( 1/2*(M+Transpose(M)) );
    n := Ncols(D);
    return DiagonalQuadraticForm(K, [ D[j,j] : j in [1..n] ]);
end intrinsic;

intrinsic Dimension( q::DiagQuadFrm ) -> RngIntElt
{
Dimension of q
}
    return q`dimension;
end intrinsic;

intrinsic BaseRing( q::DiagQuadFrm ) -> Rng
{
Returns the base ring of q
}
    return q`base_ring;
end intrinsic;

intrinsic BaseField( q::DiagQuadFrm ) -> Rng
{ " } //"
    return q`base_ring;
end intrinsic;

intrinsic Discriminant( q::DiagQuadFrm ) -> RngElt
{
Discriminant of a quadratic form q.
}
    if not assigned q`disc then
        d := Dimension(q);
        c := Integers() ! ((-1)^(d*(d-1)/2));
        q`disc := (q`base_ring ! c)* &* q`coeffs;
    end if;
    return q`disc;
end intrinsic;

intrinsic IsNondegenerate( q::DiagQuadFrm ) -> BoolElt
{
Return true iff the form q is non-degenerate i.e. its determinant is non-zero.
}
    return forall{ c : c in q`coeffs | not IsZero(c) };
end intrinsic;

intrinsic IsNondegenerate( V::ModTupFld ) -> BoolElt
{ " } //"
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";

    return not IsZero(Determinant(V));
end intrinsic;

intrinsic OrthogonalSum( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{
Orthogonal sum of two diagonal quadratic forms.
}
    require q1`base_ring eq q2`base_ring: "Forms must be defined over the same ring.";
    
    return DiagonalQuadraticForm( q1`base_ring, q1`coeffs cat q2`coeffs );
end intrinsic;

intrinsic '+'( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{ " } //"
    return OrthogonalSum(q1, q2);
end intrinsic;

intrinsic '-'( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{
Computes an ``orthogonal difference'', i.e. a sum q1 + (-q2)
}
    return OrthogonalSum(q1, DiagonalQuadraticForm( q2`base_ring, [ -c : c in q2`coeffs ]));
end intrinsic;

intrinsic TensorProduct( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{
Tensor product of two diagonal quadratic forms.
}
    require q1`base_ring eq q2`base_ring: "Forms must be defined over the same ring.";
    
    return DiagonalQuadraticForm( q1`base_ring, [ c*d : c in q1`coeffs, d in q2`coeffs ] );
end intrinsic;

// constructor of a Pfister form
intrinsic PfisterForm( K::Rng, D::[RngElt] ) -> DiagQuadFrm
{
Given a sequence of coefficients D = [d1, d2,..., dn], constructs 
a Pfister form q = ( 1, d1, d2, ..., dn, d1*d2, d1*d3, ..., d1*d2*...*dn ).
}
    require forall{ d : d in D | IsCoercible(K, d) }: "Coefficients must be coercible to the base field.";
    q := DiagonalQuadraticForm( K, [One(K), D[1]] );
    for j in [2..#D] do
        q1 := DiagonalQuadraticForm( K, [One(K), D[j]] );
        q := TensorProduct( q, q1 );
    end for;
    return q;
end intrinsic;

intrinsic PfisterForm( D::[RngElt] ) -> DiagQuadFrm
{ " } //"
    PfisterForm( Universe(D), D );
end intrinsic;

intrinsic RandomShuffle( q::DiagQuadFrm : n := 3 ) -> DiagQuadFrm
{
Given a diagonal quadratic form q, this function creates a form q1, which is 2-conneced to q in n steps. By default n equals 3, but can be set to any positive integer.
}
    d := Dimension(q);
    C := q`coeffs;
    require d ge 2: "The dimension of the form must be at least 2.";
    for j in [1..n] do
        // pick two random coefficients
        repeat
            i1 := Random(1, d);
            i2 := Random(1, d);
        until i2 ne i1;
        a := C[i1];
        b := C[i2];
        a_ := a + b;
        if IsZero(a_) then
            continue;
        end if;
        b_ := a*b*a_;
        C[i1] := a_;
        C[i2] := b_;
    end for;

    return DiagonalQuadraticForm( BaseRing(q), C );
end intrinsic;

intrinsic Coefficients( q::DiagQuadFrm ) -> []
{
Returns the sequence of the coefficients of q.
}
    return q`coeffs;
end intrinsic;
