freeze;

import "quadratic_forms.m" : signatures_builtin, primes_that_matter;
import "../helpers/varia.m" : ZZ, FF2;

// helper structure to iterate over primes of K
PRIME_ITERATOR := recformat< p : Integers(), j : Integers(), f, K >;

function initialize_prime_iterator( K )
// This function initializes an iterator for non-dyadic primes of K
// It returns a PRIME_ITERATOR object
    return rec< PRIME_ITERATOR | p := 2, j := 0, f := [], K := K >;
end function;

function next_prime( i )
// Given a PRIME_ITERATOR object i, this function returns 
// "the next" prime of K.
    if i`j ge #i`f then
        i`p := NextPrime(i`p);
        i`f := [ Ideal(P[1]) : P in Decomposition(i`K, i`p) ];
        i`j := 0;
    end if;
    i`j +:= 1;
    return i, i`f[i`j]; 
end function;


// Ordering separation functions

function strong_ordering_separation_helper(a, b, S)
// Given too real numbers a < b and a finite set S of
// prime ideals of a number field K, this function outputs a rational
// number α ∈ (a, b) such that α is a local square at every place p ∈ S.
// [ KR21, Algorithm 2 ]

    assert (a lt b);
    
    // rational primes dominated by places in S
    // strange syntax to make elements unique
    S_ := Setseq({@  MinimalInteger(s) : s in S @});
    if 2 in S_ then 
        s := 4 * (&* S_);
    else
        s := &* S_;
//        m := ChineseRemainderTheorem( [1 : p in S_], S_)
    end if;
    l := 1 + Ceiling( 1/ (b-a) - 1/s );
    k := Ceiling( (a*(1 + l*s) - 1)/s );
    if not ( a lt (1+k*s)/(1+l*s) ) then
        k +:= 1;
    end if;
    assert (1+k*s)/(1+l*s) lt b;
    assert a lt (1+k*s)/(1+l*s);
    return (1+k*s)/(1+l*s);
end function; 

function strong_ordering_separation_helper_v2(a, b, S)
// Given too real numbers a < b and a finite set S of
// prime ideals of a number field K, this function outputs a rational
// number α ∈ (a, b) such that α is a local square at every place p ∈ S.
// Brute-force approach. Hope it will get nicer results

    assert (a lt b);
    
    // find n s.t. (b-a) > 1/2^n
    n := 1 + Ceiling(Log(2, 1/(b-a)));

    flag := true;
    while flag do
        k := Ceiling(a*2^n);
        while k/2^n lt b do
            if forall{ s : s in S | IsLocalSquare( k/2^n, s ) } then
                alpha := k/2^n;
                flag := false;
                break;
            else
                k +:= 1;
            end if;
        end while;
        n +:= 1;
    end while;

    assert (alpha gt a) and (alpha lt b);
    assert forall{ s : s in S | IsLocalSquare( alpha, s ) };
    return alpha;
end function; 



function positive_approximation( S, L )
// Given a finite set S = {p1, . . . , pn } of non-archimedean 
// primes and some elements λ1, . . . , λn ∈ K, this function constructs a
// *totally positive* element α ∈ K such that α ≡ λi ( mod pi^2 )
// for every i ≤ n.
// This is an implementation of [KR202?, Algorithm 6]

    // the base field
    K := NumberField(Universe(L));

    beta := ChineseRemainderTheorem( L, [ p^2 : p in S ] );
    // rational primes dominated by places in S
    // strange syntax to make elements unique
    S_ := Setseq({@  MinimalInteger(s) : s in S @});
    s := &* S_;
    
    n := 1 + Ceiling( Max([ Evaluate(-beta, sigma) : sigma in RealPlaces(K) ])/s^2 );
    
    return beta + n*s^2;
end function;



function strong_ordering_separation_v1( I, S )
// Given a list I of signs at real places and a finite set S
// of prime ideals, this function constructs rho s.t.
// rho is negative at places in I and positive in other real places
// rho is a square in each place in S

    // the base field
    K := NumberField(Universe(I));
    theta := K.1;
    r, _ := Signature(K);
    embeddings := RealPlaces(K);
    evals      := [ Evaluate(theta, s) : s in embeddings ];
    ParallelSort( ~evals, ~embeddings );

    // rational primes dominated by places in S
    // strange syntax to make elements unique
    S_ := Setseq({@  MinimalInteger(s) : s in S @});
    if 2 in S_ then 
        s := 4 * (&* S_);
    else
        s := &* S_;
    end if;
    // the definining poly and all its real roots
    // WARNING: floating point numbers are used here
    f := DefiningPolynomial(K);
    M := Ceiling( s*RealRootBound(f) );
    XX := [ -M ] cat [ s*xx[1] : xx in Roots(f, RealField()) ] cat [ M ];
    // elements in-between that are local squares in S
    A := [ strong_ordering_separation_helper_v2( XX[i], XX[i+1], S ) : i in [1..#XX-1] ];
    
    Etas := [ (s*theta - A[i])*(s*theta - A[i+1]) : i in [1..#A-1] ];
    //OK := MaximalOrder(K);
    //print [ Factorization(eta*OK) : eta in Etas ];
    //alpha := &* ([1] cat [ Etas[Index(embeddings, s)] : s in I ]);
    alpha := ProductRepresentation(
        [K | Etas[s] : s in [1..r] | I[s] eq -1 ], 
        [ZZ | 1 : s in [1..r] | I[s] eq -1 ]);
    return alpha;
end function; 

function strong_ordering_separation_v2( I, S )
// Given a list of real places I and a finite set S
// of prime ideals, this function constructs rho s.t.
// rho is negative at places in I and positive in other real places
// rho is a square in each place in S
// This version uses built-in method ChineseRemainderTheorem

    // the base field
    K := NumberField(Universe(I));
    OK := MaximalOrder(K);
    theta := K.1;
    r, _ := Signature(K);
    embeddings := RealPlaces(K);
    
    // ideal made out of S
    s := &* S;
    alpha := ChineseRemainderTheorem(4*s, [1..#embeddings], OK!1, I);
    
    assert forall{ P : P in S | IsLocalSquare(alpha, P) };
    assert forall{ sigma : sigma in I | Evaluate(alpha, sigma) lt 0 };
    assert forall{ sigma : sigma in [ sig : sig in embeddings | sig notin I] | Evaluate(alpha, sigma) gt 0 };
    return alpha;
end function; 

function strong_ordering_separation_v3( I, S )
// Given a list of real places I and a finite set S
// of prime ideals, this function constructs rho s.t.
// rho is negative at places in I and positive in other real places
// rho is a square in each place in S
// New version using weak separation and positive approximation

    // the base field
    K := NumberField(Universe(I));
    OK := MaximalOrder(K);
    if not IsAbsoluteField(K) then
        K := AbsoluteField(K);
    end if;
    
    try 
        // this may fail
        alpha1 := OrderingSeparation(I : algorithm := "randomized");
    catch e
        // fall back on the deterministic
        print "Falling back to a deterministic ordering separation. It may take a long time.";
        alpha1 := OrderingSeparation(I : algorithm := "deterministic");
    end try;
    
    vprint DQF_verbosity : "alpha1 =", alpha1;
    
    // alpha2 must be totally positive and congruent to alpha1
    // in every place of S
    alpha2 := positive_approximation( S, [ OK!Numerator(alpha1) : p in S ] );
    alpha2 *:= positive_approximation( S, [ OK!Denominator(alpha1) : p in S ] );
    vprint DQF_verbosity : "alpha2 =", alpha2;
    
    return alpha1*alpha2;

end function;


// Anistropic part

function NF_anisotropic_part_dim4( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension d ≥ 4, 
// this function constructs an element α ∈ K × such that dim_a( q ⊥ <−α>) = d − 1.

    K := BaseField(q);
    if not IsReal(K) then
        return 1;
    end if;
    
    d := AnisotropicDimension(q);
    // find the real embedding where the dimension is -d
    r, _ := Signature(K);
    embeddings := RealPlaces(K);
    S := signatures_builtin(q);
    I := [ ZZ | S[i] eq -d select -1 else 1 : i in [1..r] ];
    // find alpha that is negative at these embeddings
    alpha := OrderingSeparation(I);
    return alpha;
end function;

function NF_anisotropic_part_dim3_nonreal( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 3, 
// this function constructs an element α ∈ K, such that dim_a( q ⊥ <−α>) = 2.
// This is a version for non-real fields
// This is an implementation of [KR202?, Algorithm 5]

    // base field
    K  := BaseField(q);
    OK := MaximalOrder(K);
    // find the primes that matter
    S  := primes_that_matter(q);
    d  := Discriminant(q);
    S1 := [ p : p in S | IsOdd(Valuation(d, p)) ];
    E1 := [ OK!d-1 : p in S1 ];
    S2 := [ p^2 : p in S | IsEven(Valuation(d, p)) ];
    E2 := [ OK!UniformizingElement(p) : p in S | IsEven(Valuation(d, p)) ];
    alpha := ChineseRemainderTheorem(E1 cat E2, S1 cat S2);
    
    return alpha;
end function;

function NF_anisotropic_part_dim3_real( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 3, 
// this function constructs an element α ∈ K, such that dim_a( q ⊥ <−α>) = 2.
// This is a version for formally real fields
// This is an implementation of [KR202?, Algorithm 7]

    // base field
    K  := BaseField(q);
    OK := MaximalOrder(K);
    // find the primes that matter
    S  := primes_that_matter(q);
    r, _ := Signature(K);
    //embeddings := RealPlaces(K);
    signs := signatures_builtin(q);
    //I := [ embeddings[j] : j in [1..#embeddings] | signatures_builtin(q)[j] lt 0 ];
    I := [ signs[j] lt 0 select -1 else 1 : j in [1..r] ];

    // alpha1 will be a local square at every p in S
    // and will be negative in every sigma in I
    alpha1 := strong_ordering_separation_v3(I, S);
    vprint DQF_verbosity : "alpha_1 =", alpha1;

    // alpha 2 will be totally positive and will be congruent to either disc(q) -1 or \pi_p
    d  := Discriminant(q);
    S1 := [ p : p in S | IsOdd(Valuation(d, p)) ];
    E1 := [ OK!d-1 : p in S1 ];
    S2 := [ p^2 : p in S | IsEven(Valuation(d, p)) ];
    E2 := [ OK!UniformizingElement(p) : p in S2 ];
    alpha2 := positive_approximation(S1 cat S2, E1 cat E2);
    vprint DQF_verbosity : "alpha_2 =", alpha2;
    
    return ProductRepresentation([ K!alpha1, K!alpha2], [1,1]);
end function;

function NF_anisotropic_part_dim2_nonreal( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 2, 
// this function constructs ist anisotropic part <a,b>.
// This is a version for non-real fields
// This is an implementation of [KR202?, Algorithm 8]
    // base field
    K  := BaseField(q);
    OK := MaximalOrder(K);
    // find the primes that matter
    S  := primes_that_matter(q);
    
    // make sure the Witt index is divisible by 4
    w := (WittIndex(q) mod 4);
    if w ne 0 then
        q := q + HyperbolicForm(K, 2*(4-w));
        q`primes := S;
    end if;
    
    d := Discriminant(q);
    
    found := false;
    // initialization of places' search
    it := initialize_prime_iterator(K);
    while not found do
        // basis of the group of S-singular elements
        // E_S group of singular elements 
        // psi : OK -> E_S
        E_S, psi := pSelmerGroup(2, Set(S) : Nice);
        B_ := [ K! (g @@ psi) : g in Generators(E_S) ];
        vprint DQF_verbosity : "basis:", B_;
    
        // vector of exponents of Hasse invariants
        W := Vector(FF2, [ZZ | (1 - HasseMinkowskiInvariant(q, P))/2 : P in S ]);
        vprint DQF_verbosity : "W =", W;
    
        // matrix of Hilbert symbols
        H := Matrix(FF2, [[ZZ | (1 - HilbertSymbol(b, d, P))/2  : b in B_ ] : P in S ]);
        vprint DQF_verbosity : "H =", H;
        
        // V solution to H*V == W
        found, V, _ := IsConsistent(Transpose(H), W);
        vprint DQF_verbosity : "V =", V;
        
        if found then
            // the sought element
            alpha := ProductRepresentation(B_, V_) where V_ := Eltseq( ChangeRing(V,ZZ));
        else
            it, P := next_prime(it);
            while P in S do
                it, P := next_prime(it);
            end while;
            Append(~S, P);
        end if;
        
    end while;
    
    return DiagonalQuadraticForm(K, [alpha, -alpha*d]);
end function;

function NF_anisotropic_part_dim2_real( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 2, 
// this function constructs ist anisotropic part <a,b>.
// This is a version for real fields
// This is an implementation of [KR202?, Algorithm 8]

    // base field
    K  := BaseField(q);
    OK := MaximalOrder(K);
    
    // find the primes that matter
    S  := primes_that_matter(q);   
    
    // make sure the Witt index is divisible by 4
    w := (WittIndex(q) mod 4);
    if w ne 0 then
        q := q + HyperbolicForm(K, 2*(4-w));
        q`primes := S;
    end if;
    d := Discriminant(q);
    
    // find the real places that matter
    r, _ := Signature(K);
    embeddings := RealPlaces(K);
    disc_signs := RealSigns(d);
    signatures := signatures_builtin(q);
    INX := [ j : j in [1..r] | disc_signs[j] lt 0 ];
    embeddings := [ embeddings[j] : j in INX ];
    
    found := false;
    // initialization of places' search
    it := initialize_prime_iterator(K);
    while not found do
        // basis of the group of S-singular elements
        // E_S group of singular elements 
        // psi : OK -> E_S
        E_S, psi := pSelmerGroup(2, Set(S) : Nice);
        B_ := [ K! (g @@ psi) : g in Generators(E_S) ];
        vprint DQF_verbosity : "basis:", B_;

        // vector of signs
        W_ := [ZZ | 1/2 - signatures[j]/4 : j in INX];
        // vector of exponents of Hasse invariants
        W_ := W_ cat [ZZ | (1 - HasseMinkowskiInvariant(q, P))/2 : P in S ];
        W := Vector(FF2, W_);
        vprint DQF_verbosity : "W =", W;
    
        // matrix of signs (already transposed => r cols x #B_ rows)
        A := Matrix([RealSigns(b) : b in B_ ]);
        A := Matrix(FF2, [[ZZ | (1-A[j,i])/2 : i in INX ] : j in [1..#B_]]);
        // matrix of Hilbert symbols (already transposed => #S cols x #B_ rows)
        B := Matrix(FF2, [[ZZ | (1 - HilbertSymbol(b, d, p))/2  : p in S ] : b in B_ ]);
        // join the two matrices
        AB := HorizontalJoin(A, B);
        vprint DQF_verbosity : "AB =", AB;
        
        // V solution to H*V == W
        found, V, _ := IsConsistent(AB, W);
        vprint DQF_verbosity : "V =", V;
        
        if found then
            // the sought element
            alpha := ProductRepresentation(B_, V_) where V_ := Eltseq( ChangeRing(V,ZZ));
        else
            it, P := next_prime(it);
            while P in S do
                it, P := next_prime(it);
            end while;
            Append(~S, P);
        end if;
        
    end while;
    
    return DiagonalQuadraticForm(K, [alpha, -alpha*d]);
end function;

function anisotropic_part_NF( q )
// Given a diagonal quadratic form q over a number field, this function computes its 
//anisotropic part q_a, that is an anisotropic form s.t. q = q_a ⊥ w*<1,-1>.

    K := BaseField(q);
    P := primes_that_matter(q);
    d := AnisotropicDimension(q);
    
    // first short-cut
    if d eq Dimension(q) then
        return q;
    end if;
    
    // initialize the anisotropic part to be an empty form
    qa := DiagonalQuadraticForm(K, []);
    
    // second short-cut
    C := Coefficients(q);
    for c in C do
        q_ := q - DiagonalQuadraticForm(K, [c]);
        // substracting an existing coeff will not change primes
        // don't waste time on a new factorization
        q_`primes := P;
        if AnisotropicDimension(q_) lt d then
            q := q_;
            q`primes := P;
            qa +:= DiagonalQuadraticForm(K, [c]);
            d -:= 1;
        end if;
    end for;
    
    while d ge 4 do
        a   := NF_anisotropic_part_dim4( q );
        a_  := DiagonalQuadraticForm(K, [a]);
        qa +:= a_;
        q  -:= a_;
        d  -:= 1;
    end while;
    
    if d eq 3 then
        if IsReal(K) then
            a   := NF_anisotropic_part_dim3_real( q );
        else 
            a   := NF_anisotropic_part_dim3_nonreal( q );
        end if;
        a_  := DiagonalQuadraticForm(K, [a]);
        qa +:= a_;
        q  -:= a_;
        d   := 2;
    end if;

    if d eq 2 then
        if IsReal(K) then
            q_ := NF_anisotropic_part_dim2_real( q );
        else
            q_ := NF_anisotropic_part_dim2_nonreal( q );
        end if;
        qa +:= q_;
    end if;
    
    if d eq 1 then
        qa +:= DiagonalQuadraticForm(K, [Discriminant(q)]);
    end if;
    
    return qa;
end function;
