freeze;

//////////////////////////////////////////////////
// IsLocalSquare

intrinsic IsLocalSquare( a::FldNumElt, P::PlcNumElt ) -> BoolElt
{ 
Given an element a in K, decide if a is a square in the completion K_P
}
    return IsLocalSquare(a, Ideal(P));
end intrinsic;


//////////////////////////////////////////////////
// IsSumOfSquares

intrinsic IsSumOfSquares( a::FldNumElt ) -> BoolElt
{
Returns true iff a is a sum of squares
}
    //return IsTotallyPositive(a);
    return forall{ s : s in RealSigns(a) | s eq 1 };
end intrinsic;

intrinsic IsSOS( a::FldNumElt ) -> BoolElt
{ " } //"
    return IsSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

function length_nondyadic( a, P )
// Given a nonzero element a of a number field K and a nondyadic place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  
    
    // the residue field
    Kp := ResidueClassField(P);
    // and valuation
    v := Valuation(a,P);
    // two cases to consider
    if IsOdd(v) then
        // possible values are 2 and 3
        q := DiagonalQuadraticForm(Kp, [Kp!1, Kp!1]);
        return IsIsotropic(q) select 2 else 3;
    else
        // possible values are 1 and 2
        return IsLocalSquare(a, P) select 1 else 2;
    end if;
end function;

function length_dyadic( a, P )
// Given a nonzero element a of a number field K and a dyadic place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  

    K := Parent(a);
    // if it is a square, the answer is trivial
    if IsLocalSquare(a, P) then
        return 1;
    end if;
    // now possible values are 2, 3, 4
    // check for 2
    if HilbertSymbol(K!-1, a, Ideal(P)) eq 1 then
        return 2;
    end if;
    // the two possibilities left are 3 and 4
    // check 3
    if HilbertSymbol(K!-1, K!-1, Ideal(P)) eq 1 or not IsLocalSquare(-a, P) then
        return 3;
    end if;
    // now the only choice is 4
    return 4;
end function;

intrinsic LengthOfSOS( a::FldNumElt ) -> RngIntElt
{
Given a nonzero element a of a number field K, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    require not IsZero(a): "The argument must be nonzero!";
    // start with trivial cases
    if not IsSOS(a) then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // the possible values now are 2, 3, 4
    K := Parent(a);
    OK := MaximalOrder(K);
    // find all the primes of K that matter for a
    even_primes := [ P[1] : P in Decomposition(K, 2) ];
    //odd_primes :=  [ P[1] : P in Factorization(a*OK) ];
    odd_primes, _ := [ P : P in Support(Divisor(a)) | P notin even_primes ];
    vprint DQF_verbosity : "Primes to be tested:", even_primes, odd_primes;
    
    vprint DQF_verbosity : "Lengths at dydadic primes:", [ length_dyadic(a, P) : P in even_primes ];
    m := Max([ length_dyadic(a, P) : P in even_primes ]);
    vprint DQF_verbosity : "m = ", m;
    vprint DQF_verbosity : "Lengths at non-dydadic primes:", [ length_dyadic(a, P) : P in odd_primes ];
    n := IsEmpty(odd_primes) select -1 else Max([ length_nondyadic(a, P) : P in odd_primes ]);
    vprint DQF_verbosity : "n =", n;
    l := Max([2, m, n]);
    return l;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldNumElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


//////////////////////////////////////////////////
// LegendreSymbol

intrinsic LegendreSymbol( a::FldNumElt, P::PlcNumElt ) -> RngIntElt
{ 
Given an element a and a place P of a number field, 
this function computes the Legendre symbol a over P. 
}
    v := Valuation(a, P);
    if IsOdd(v) then
        return 0;
    end if;
    if v ne 0 then
        p := UniformizingElement(P);
        a := a*p^(-v);
    end if;
    if IsSquare(Evaluate(a,P)) then
        return 1;
    end if;
    return -1;
end intrinsic;

intrinsic LegendreSymbol( a::FldNumElt, P::RngOrdIdl ) -> RngIntElt
{ " }
    require IsPrime(P): "P must be a prime ideal!";
    return LegendreSymbol(a, Place(P));
end intrinsic;


