freeze;

import "../helpers/varia.m" : ZZ, FF2;

declare attributes FldNum: Lvl, PythNum, PythNumElt, WittEqInv, Hght, SortedRealPlaces, Etas;

WittEqInv_fmt := recformat<
    d,  // absolute degree`
    r,  // number of real embeddings
    s,  // level
    k,  // number of dyadic places
    L   // list of pairs (degree, level) for all dyadic places
>;

//////////////////////////////////////////////////
// IsFormallyReal

intrinsic IsFormallyReal( K::FldNum ) -> BoolElt
{
Returns true iff K is formally real (i.e. -1 is not a sum of squares)
}
    r,c := Signature(K);
    return r ne 0;
end intrinsic;

intrinsic IsReal( K::FldNum ) -> BoolElt
{ " } //"
    return IsFormallyReal(K);
end intrinsic;





//////////////////////////////////////////////////
// Level

intrinsic Level( K::FldNum ) -> RngIntElt
{    
This function comutes the level s(K) of a number field K,
i.e. a lowest number s such that -1 is a sum of s squares
it returns 0 if K is formally real.
}
    if assigned K`Lvl then
        return K`Lvl;
    end if;

    // level of real field is 0 (or infinity)
    if IsFormallyReal(K) then
        K`Lvl := 0;
        return 0;
    end if;
    // is level equal 1
    if IsSquare(K!-1) then
        K`Lvl := 1;
        return 1;
    end if;
    // now level is either 2 or 4 depending on dyadic primes
    D := Decomposition(K, 2);
    for d in D do
        f := InertiaDegree(d[1]);
        e := RamificationIndex(d[1]);
        if IsOdd(f) and IsOdd(e) then
            K`Lvl := 4;
            return 4;
        end if;
    end for;
    K`Lvl := 2;
    return 2;
end intrinsic;





//////////////////////////////////////////////////
// OrderingSeparation

function random_vector( n : min_range := -2^10,  max_range := 2^10)
// given a dimension n, this function constructs a random vector
// with integer coefficients.
    return Vector(ZZ, [ Random(min_range, max_range) : j in [1..n] ]);
end function;

function ordering_separation_randomized( K, I : max_count := 1000 )
// Given a set or sequence of indices of real places, this function constructs an
// element that is negative precisely at the selected places and positive 
// in all other places.
// I must be a sequence of signs +/-1 for each ordering of K

    // the trivial cases
    if forall{ s : s in I | s eq 1} then
        return One(K);
    elif forall{ s : s in I | s eq -1} then
        return -One(K);
    end if;
    
    // vector of the signs we want
    W  := Vector(FF2, [ZZ | (1-s)/2 : s in I]);
    W_ := Vector(FF2, [ZZ | 1-(1-s)/2 : s in I]);
    
    f := DefiningPolynomial(K);
    r, _ := Signature(K);
    // initial amplitude of distortion
    m := 4;
    // build the sequence of derivatives as the initial guess
    g := f;
    L := [Parent(f)| ];
    while Degree(g) ge 0 do
        if not IsZero(Resultant(f,g)) then 
            Append(~L, g);
        else
            // if f and g have a common root then shifting g up and down
            // we may likely isolate a single root of f
            j := 0;
            while IsZero(Resultant(f,g + j)) do
                j +:= 1;
            end while;
            Append(~L, g+j);
            j := 0;
            while IsZero(Resultant(f,g + j)) do
                j -:= 1;
            end while;
            Append(~L, g+j);
        end if;
        g := Derivative(g);
    end while;
    //L := [ Derivative(f,j) : j in [1..Degree(f)] ];

    // flag to stop the loop
    found := false;
    alpha := 1;
    counter := 1;
    
    while not found do
        // evaluate the polys at theta
        Lt := [ Evaluate(g, K.1) : g in L ];
        
        // matrix of signs of the polynomials 
        C_ := Matrix(ZZ, [RealSigns(a) : a in Lt ]);
        C := Matrix(FF2, [[ZZ | (1-C_[i,j])/2 : j in [1..Ncols(C_)]] : i in [1..Nrows(C_)]]);

        // try to solve
        found, V, _ := IsConsistent(C, W);
        if found then
            //print "V =", V;
            alpha := ProductRepresentation([K | a : a in Lt], V_) 
                where V_ := Eltseq( ChangeRing(V,ZZ));
        else
            found, V, _ := IsConsistent(C, W);
            if found then
                alpha := ProductRepresentation([K | a : a in Lt], V_) 
                    where V_ := Eltseq( ChangeRing(V,ZZ));
            end if;
        end if;
        // increase the amplitude of distrotion
        counter +:= 1;
        // add a new polynomial
        g := Parent(f)!Eltseq(random_vector(Degree(f)-1 : min_range := -m, max_range := m ));
        Append( ~L, g );
        
        if counter mod 10 eq 0 then
            m *:= 2;
        end if;
        if counter gt max_count then
            return false, 0;
        end if;
    end while;
    
    return true, alpha;
end function;

function ordering_separation_deterministic( K, I )
// Given a set or sequence of indices of real places, this function constructs an
// element that is negative precisely at the selected places and positive 
// in all other places.
// I must be a sequence of signs +/-1 for each embeddings.

    // the trivial cases
    if forall{ s : s in I | s eq 1} then
        return One(K);
    elif forall{ s : s in I | s eq -1} then
        return -One(K);
    end if;

    r, _ := Signature(K);
    
    if assigned K`Etas then
        Etas := K`Etas;
    else
        // find etas that isolate each root of f in an ascending order
        theta      := K.1;
        f          := DefiningPolynomial(K);
        Etas := [ (theta - ab[1])*(theta - ab[2]) : ab in AB ]
            where AB is RealRootIntervals(f);
        // sort the real embeddings of K accordingly
        embeddings := RealPlaces(K);
        evals      := [ Evaluate(theta, s) : s in embeddings ];
        indices    := [ 1..r ];
        ParallelSort( ~evals, ~indices );
        K`Etas := [ Etas[Index(indices,j)] : j in [ 1..r ] ];
        Etas := K`Etas;
    end if;
    
    if #[s : s in I | s eq -1] le r/2 then
        alpha := &* ([1] cat [ Etas[s] : s in [1..r] | I[s] eq -1 ]);
    else
        alpha := - &* ([1] cat [ Etas[s] : s in [1..r] | I[s] eq 1 ]);
    end if;
    
    // make alpha integral
    if Denominator(alpha) in ZZ then
        alpha := Sign(Denominator(alpha))*Numerator(alpha);
    end if;
    
    return alpha;
end function;


intrinsic OrderingSeparation( 
    K::FldNum, I::[ RngIntElt ] : algorithm := "deterministic", max_count := 1000 
) -> FldNumElt
{
Given a set or sequence of signs +/-1 one for each real place of a number field, 
this function constructs an element that is has specified signs.
}
    require IsAbsoluteField(K) : "The number field must be an absolute extension of QQ.";
    r, _ := Signature(K);
    require #I eq r : "Length of the second argument does not match the number of real places of the field.";
    require forall{s : s in I | (s eq 1) or (s eq -1)} : "All the entries in the second argument must be equal either 1 or -1.";
    require algorithm in ["randomized", "deterministic"] : "The only allowed values for the parameter \'algorithm\' are \"randomized\" and \"deterministic\".";
    
    if algorithm eq "randomized" then
        found, a := ordering_separation_randomized(K, I : max_count := max_count);
        error if not found, "Maximal number of iterations exceeded!";
        return a;
    else 
        return ordering_separation_deterministic(K, I);
    end if;
end intrinsic;


//////////////////////////////////////////////////
// PythagorasNumber & PythagorasElement

intrinsic PythagorasNumber( K::FldNum ) -> RngIntElt
{
This function computes the Pythagoras number P(K) of a number field K,
i.e. the lowest number p such that every sum of squares in K
is a sum of p squares.
}
    if assigned K`PythNum then
        return K`PythNum;
    end if;

    // if s(K)=1 then P(K)=2
    if IsSquare(K!-1) then
        K`PythNum := 2;
        return K`PythNum;
    end if;
    // now level is either 3 or 4 depending on dyadic primes
    D := Decomposition(K, 2);
    for d in D do
        f := InertiaDegree(d[1]);
        e := RamificationIndex(d[1]);
        if IsOdd(f) and IsOdd(e) then
            K`PythNum := 4;
            return K`PythNum;
        end if;
    end for;
    K`PythNum := 3;
    return K`PythNum;
end intrinsic;

intrinsic PythagorasElement( K::FldNum ) -> FldNumElt
{
Returns an element a of K whose length equals the Pythagoras 
number P(K) of K.
}
    if assigned K`PythNumElt then
        return K`PythNumElt;
    end if;
    
    PK := PythagorasNumber(K);
    sK := Level(K);
    if PK eq 2 then
        // K is nonreal by [Lam 2005, Exm. XI.5.9] 
        // and its level must be 1 by [Koprowski, Czogała 2018, Prop. 4]
        a := K!-1;
        while not assigned K`PythNumElt do
            c, a_ := IsSquare(a);
            if not c then 
                K`PythNumElt := a;
            else
                a := a_;
            end if;
        end while;
    elif PK eq 3 then
        // Level is 0 or 2
        p := 2;
        while not assigned K`PythNumElt do
            D := Decomposition(K, p);
            for P in D do
                if not IsSquare( Evaluate(K!-1, P[1]) ) then
                    // this is ad hoc correction, more work needed
                    a := UniformizingElement( P[1] );
                    if IsSOS(a) then 
                        K`PythNumElt := UniformizingElement( P[1] );
                        break;
                    end if;
                end if;
            end for;
            p := NextPrime(p);
        end while;
    elif PK eq 4 then
        // level is either 0 or 4
        K`PythNumElt := K!7;
    else
        error if false, "Something went wrong!";
    end if;
    assert2 LengthOfSOS(K`PythNumElt) eq PythagorasNumber(K);
    return K`PythNumElt;
end intrinsic;





//////////////////////////////////////////////////
// Height

intrinsic Height( K::FldNum ) -> RngIntElt
{
This function computes the height h(K) of a number field K,
i.e. exp( Wt(K) ), where Wt(K) is denotes the torsion subgroup
of the Witt ring of K.
}
    if assigned K`Hght then
        return K`Hght;
    end if;
    
    if not IsFormallyReal(K) then
        // if K is non-real, then h(K) = 2*s(K)
        // see [Lam, p. 395]
        K`Hght := 2*Level(K);
    else
        // otherwise it is 4 by
        // [Lam, Thm 5.6] and posible values of P(K)
        K`Hght := 4;
    end if;
    return K`Hght;
end intrinsic;





//////////////////////////////////////////////////
// AreWittEquivalent

function local_dyadic_level( d )
    // For a dyadic place d of a number field K, 
    // this function computes the level s(Kd) of 
    // the completion Kd of K at d
    
    K := NumberField(d);
    // if s(K)=1 then s(Kd) = 1
    if IsSquare(K!-1) then
        return 1;
    end if;
    
    f := InertiaDegree(d);
    e := RamificationIndex(d);
    if IsOdd(e) and IsOdd(f) then
        return 4;
    end if;
    if IsOdd(e) and IsEven(f) then
        return 2;
    end if;
    I := Ideal(d);
    R := Order(I);
    Kd, phi := Completion(R, I);
    if IsSquare(phi(-1)) then
        return 1;
    end if;
    return 2;
end function;

function Witt_eq_invariant( K )
    // Given a number field K, this function computes its
    // complete set of Witt equivalence invariants
    
    if not assigned K`WittEqInv then
        r, _ := Signature(K);
        D := Decomposition(K, 2);
        L := [ <LocalDegree(d), local_dyadic_level(d)> where d := d_[1] : d_ in D ]; 
        Sort(~L);
        K`WittEqInv := rec< WittEqInv_fmt | 
            d := AbsoluteDegree(K),
            r := r,
            s := Level(K),
            k := #D,
            L := L
        >;
    end if;
    return <K`WittEqInv`d, K`WittEqInv`r, K`WittEqInv`s, K`WittEqInv`k, K`WittEqInv`L>;
end function;

intrinsic AreWittEquivalent( K::FldNum, L::FldNum ) -> BoolElt
{
Given two fields K and L, this function return true if their Witt rings are isomorphic and false otherwise.
}    
    WK := Witt_eq_invariant(K);
    WL := Witt_eq_invariant(L);
    return WK eq WL;
end intrinsic;

intrinsic AreWittEquivalent( K::FldNum, L::FldRat ) -> BoolElt
{ " } //"
    return AbsoluteDegree(K) eq 1;
end intrinsic;

intrinsic AreWittEquivalent( K::FldNum, L::FldFun[FldFin] ) -> BoolElt
{ " } //"
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldNum, L::FldFunRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

