freeze;

import "../helpers/varia.m" : ZZ, QQ;

declare attributes DiagQuadFrm: primes;


//////////////////////////////////////////////////
// HasseMinkowskiInvariant
// WittInvariant

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::RngOrdIdl ) -> RngIntElt
{
The Hasse invariant of a diagonal quadratic form q = < a1, ..., ad> \
at a prime P is a product of Hilbert symbols (c.f. [Lam (2005), Def. V.3.17]) \
prod_(i<j) (ai, aj)_p.
This function computes the Hasse symbol of q at p.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";

    d := Dimension(q);
    L := [ HilbertSymbol(q`coeffs[i],q`coeffs[j], P) : j in [(i+1)..d] , i in [1..d-1] ];
    return &* L;
end intrinsic;

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::PlcNumElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, Ideal(P));
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::RngOrdIdl ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, P);
end intrinsic;
    
intrinsic WittInvariant( q::DiagQuadFrm, P::PlcNumElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, Ideal(P));
end intrinsic;


//////////////////////////////////////////////////
// primes_that_matter

function primes_that_matter( q )
    // find all the primes of K that matter for q
    if not assigned q`primes then
        K := BaseRing(q);
        OK := MaximalOrder(K);

        //even_primes := [ Ideal(P[1]) : P in Decomposition(K, 2) ];
        primes := &join [ {a[1] : a in Factorization(c*OK)} : c in (q`coeffs cat [K!2]) ];
        //odd_primes := [ P : P in odd_primes | P notin even_primes ];
        //q`primes := even_primes cat odd_primes;
        q`primes := Setseq( primes );
    end if;
    return q`primes;
end function;

//////////////////////////////////////////////////
// IsIsotropic

function is_locally_isotropic_nondyadic( q, P )
// Given a quadratic form q over a number field or QQ and a prime ideal P
// this function checks if the localization of q at p is isotropic.

    // unary forms cannot be isotropic
    if Dimension(q) le 1 then
        return false;
    end if;
    // quintic and higher dimensional forms are always isotropic
    if Dimension(q) ge 5 then
        return true;
    end if;

    // the base field and its maximal order
    K := q`base_ring;
    OK := MaximalOrder(K);
  
    // the residue field
    Kp, kappa := ResidueClassField(P);
    // uniformizer
    if ISA(Category(OK), RngInt) then
        p := Generator(P);
    else 
        p := UniformizingElement(P);
    end if;

    // split the form into q = q0 + p*q1
    q0 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsEven(Valuation(a,P)) ];
    q1 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsOdd(Valuation(a,P)) ];
    q0 := [ a*Denominator(a)^2 : a in q0 ];
    q1 := [ a*Denominator(a)^2 : a in q1 ];
    
    // take the reduction mod gp
    q0bar := [ kappa(a) : a in q0 ];
    q1bar := [ kappa(a) : a in q1 ];
    // verify if q0, q1 are isotropic
    A := IsIsotropic( DiagonalQuadraticForm(Kp, q0bar) );
    B := IsIsotropic( DiagonalQuadraticForm(Kp, q1bar) );
    return A or B;
end function;

function is_locally_isotropic_dyadic( q, P )
// Given a quadratic form q over a number field and a prime ideal P
// this function checks if the localization of q at p is isotropic.

    // the base field and its maximal order
    K := q`base_ring;

    // the dimension
    d := Dimension(q);
    // the determinant
    detq := &* q`coeffs;
    //Kd, phi := Completion(Order(P), P);
    
    if d le 1 then
        return false;
    elif d ge 5 then
        return true;
    elif d eq 2 then
        return IsLocalSquare(-detq, P);
    elif d eq 3 then
        // use [Lam, Prop. V.3.22]
        s := HilbertSymbol(K!-1, -detq, P);
        sq := WittInvariant(q, P);
        return (s eq sq);
    elif IsLocalSquare(detq, P) then
        // if we get here, then d == 4        
        // for quartic form with determinant 1, use [ Lam, Proposition V.3.23]
        s := HilbertSymbol(K!-1, K!-1, P);
        sq := WittInvariant(q, P);
        return (s eq sq);
    end if;
    // form is unique with det a square
    return true;
end function;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngOrdIdl ) -> BoolElt
{
Checks if a quadratic form q over a number field is locally isotropic 
at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    K := BaseField(q);
    if (K!2) notin P then
        return is_locally_isotropic_nondyadic(q,P);
    else
        return is_locally_isotropic_dyadic(q,P);
    end if;
end intrinsic;

intrinsic IsIsotropic( q::DiagQuadFrm, P::PlcNumElt ) -> BoolElt
{ " } //"
    return IsIsotropic(q, Ideal(P));
end intrinsic;

function signatures_builtin( q )
    // Given a quadratic form q over a number field K, this function returns
    // a sequence of signatures of q at all real embedings of K.
    // This implementation uses a built-in function RealSigns.
    
    K := BaseField(q);
    assert ISA(Category(K), FldNum);
    
    r, _ := Signature(K);
    S := [ RealSigns(c) : c in q`coeffs ];
    return [ &+ [ s[j] : s in S ] : j in [1..r] ];
end function;

function is_isotropic_NF( q )
    // Given a quadratic form q with coefficients in a number field K,
    // this function tests if q is isotropic.

    // the base field and its maximal order
    K := BaseField(q);
    OK := MaximalOrder(K);

    // the dimension
    d := Dimension(q);
    if d le 1 then
        return false;
    elif d eq 2 then
        return IsSquare(Discriminant(q));
    end if;

    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    vprint DQF_verbosity : "Signatures: ", S;
    // check if q is isotropic at every real place
    if not &and [ Abs(s) lt d : s in S ] then
        return false;
    end if;
    vprint DQF_verbosity : "Signatures passed";
          
//     // find all the primes of K that matter for q
//     if not assigned q`primes then
//         even_primes := [ Ideal(P[1]) : P in Decomposition(K, 2) ];
//         odd_primes := &join [ {a[1] : a in Factorization(c*OK)} : c in q`coeffs ];
//         odd_primes := [ P : P in odd_primes ];
//         q`primes := even_primes cat odd_primes;
//     end if;
    Pq := primes_that_matter(q);
    vprint DQF_verbosity : "Primes to be tested:", Pq;
    // check that q is locally isotropic over all of them
    return forall{ P : P in Pq | IsIsotropic(q, P) };
end function;


//////////////////////////////////////////////////
// IsHyperbolic

intrinsic IsHyperbolic( q::DiagQuadFrm, P::PlcNumElt ) -> BoolElt
{
Given a quadratic form q over a number field K and a prime P of K, 
this function returns true iff q is hyperbolic in the completion
of K at P.
}
    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsLocalSquare(Discriminant(q), P) then
        return false;
    end if;
    m := ZZ ! (d/2);
    K := q`base_ring;
    t := HilbertSymbol(K!-1,K!-1,Ideal(P))^( ZZ ! (m^2-m)/2 );
    return HasseMinkowskiInvariant(q, P) eq t;
end intrinsic;

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngOrdIdl ) -> BoolElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    return IsHyperbolic(q, Place(P));
end intrinsic;

function is_hyperbolic_NF( q )
    // Given a quadratic form q over a number field K, 
    // this function returns true if q is hyperbolic and false otherwise.
    
    // the base field and its maximal order
    K := q`base_ring;
    OK := MaximalOrder(K);

    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsSquare(Discriminant(q)) then
        return false;
    end if;

    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    // if it contains a nonzero, then q is non-hyperbolic at this embedding
    if exists{s : s in S | not IsZero(s) } then
        return false;
    end if;

    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         even_primes := [ Ideal(P[1]) : P in Decomposition(K, 2) ];
//         odd_primes := &join [ {a[1] : a in Factorization(c*OK)} : c in q`coeffs ];
//         odd_primes := [ P : P in odd_primes ];
//         q`primes := even_primes cat odd_primes;
//     end if;
    Pq := primes_that_matter(q);
    vprint DQF_verbosity : "Primes to be tested:", Pq;
    
    // check that q is locally hyperbolic over all of them
    return forall{ P : P in Pq | IsHyperbolic(q, P) };
end function;


//////////////////////////////////////////////////
// AnisotropicDimension

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngOrdIdl ) -> RngIntElt
{
Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. The form q_1 is called an anisotropic part of q.

Given a quadratic form q and a prime P, this function returns a dimension 
of an anisotropic part of localilization of q at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    
    d := Dimension(q);
    if IsEven(d) then
        // dimension of hyperbolic form is 0
        if IsHyperbolic(q, P) then
            return 0;
        elif IsLocalSquare(Discriminant(q), P) then
            return 4;
        else 
            return 2;
        end if;
    end if;
    // when q is odd-dimensional, extend it to the next even dimension
    c := QQ ! (-1)^(ZZ ! (d*(d+1)/2));
    q2 := DiagonalQuadraticForm( q`base_ring, q`coeffs cat [ c * &* q`coeffs ] );
    if IsHyperbolic( q2, P ) then
        return 1;
    else
        return 3;
    end if;
end intrinsic;

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::PlcNumElt ) -> RngIntElt
{ " } // "
    return AnisotropicDimension(q, Ideal(P));
end intrinsic;

function anisotropic_dimension_NF( q )
// Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
// as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. 
// The form q_1 is called an anisotropic part of q.
//
// Given a quadratic form q over a number field K, 
// this function returns a dimension of an anisotropic part of q.

    // the base field and its maximal order
    K := q`base_ring;
    OK := MaximalOrder(K);

    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    vprint DQF_verbosity : "Signatures: ", S;
    if not IsEmpty(S) then
        N := Max([Abs(s) : s in S]);
    else
        N := 0;
    end if;

    // for n > 2, it must be a real place, where the the anisotropic
    // part has a maximal dimension
    if N  ge 3 then
        return N;
    end if;
    vprint DQF_verbosity : "Signatures passed";
          
    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         even_primes := [ Ideal(P[1]) : P in Decomposition(K, 2) ];
//         odd_primes := &join [ {a[1] : a in Factorization(c*OK)} : c in q`coeffs ];
//         odd_primes := [ P : P in odd_primes ];
//         q`primes := even_primes cat odd_primes;
//     end if;
    Pq := primes_that_matter(q);
    vprint DQF_verbosity : "Primes to be tested:", Pq;
    // compute maximal anisotropic dimension over all primes
    M :=  Max([ AnisotropicDimension(q, P) : P in Pq ]);
    return Max(M, N);    
end function;


//////////////////////////////////////////////////
// IsTorsion

function is_torsion_NF( _q )
// Check if the element _q of a Witt ring is torsion.

    q := _q`form;
    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    // use Szymiczek (1997), Cor. 17.3.1 and Cor. 19.2.3
    return forall{ IsZero(s) : s in S };
end function;

