freeze;

import "quadratic_forms.m" : primes_that_matter;
import "../helpers/varia.m" : ZZ, FF2;

function simplify( a ) 
    a_ := Numerator(a)*Denominator(a);
    G := SquareFreeFactorization(a_);
    vprint DQF_verbosity, 2 :  "G =", G;
    r := LeadingCoefficient(a_) * &* ([1] cat [ g[1] : g in G | IsOdd(g[2]) ]);
    assert IsSquare(a*r);
    return r;
end function;

function RFF_anisotropic_part_dim3( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 3, 
// this function constructs an element α ∈ QQ, such that dim_a( q ⊥ <−α>) = 2.
// This is an implementation of [KR202?, Algorithm 7]
/*
    K := BaseField(q);
    R := RingOfIntegers(K);

    // find the primes that matter
    S  := primes_that_matter(q);
    S := [ p : p in S | IsFinite(p) ];
    d  := Discriminant(q);
    
    // alpha must be congruent to either d-1 (mod p) or p (mod p^2)
    S1 := [ Minimum(p) : p in S | IsOdd(Valuation(d, p)) ];
    print "S1 =", S1;
    E1 := [ R!(d-1) : p in S1 ];
    print "E1 =", E1;
    S2 := [ Minimum(p)^2 : p in S | IsEven(Valuation(d, p)) ];
    print "S2 =", S2;
    E2 := [ Minimum(p) : p in S2 ];
    print "E2 =", E2;
    alpha := ChineseRemainderTheorem(E1 cat E2, S1 cat S2);
    
    print "alpha =", alpha;
    
    return simplify(alpha);
*/

    // base field
    K  := BaseField(q);
    // find the primes that matter
    PP  := primes_that_matter(q);
    
    // generators of places that matter
    B := [ K!Minimum(p) : p in PP  ];
    
    J := [ j : j in [1..#PP] | AnisotropicDimension(q, PP[j]) eq 3 ];
    vprint DQF_verbosity :  "J =", J;
    V := VectorSpace(FF2, #B);
    Vset := Set(V);

    // loop over these places
    for j in J do
        // for each place find the corresponding *alternative*
        // of congruences
        p := PP[j];
        la := Minimum(p);
        
        M := Matrix(FF2, [
            [ (LegendreSymbol(K!a,p) eq -1) select 1 else 0 : a in B ],
            [ Valuation(a,p) : a in B ]
        ]);
        // the discriminant of q will drop the dimension
        W := Vector(FF2, [
            (LegendreSymbol(K!d,p) + LegendreSymbol(K!d*la,p) eq 1) select 0 else 1,
            Valuation(d, p)
        ]) where d := Discriminant(q);

        // find the *set* of solutions
        found, v0, V0 := IsConsistent( Transpose(M), W );
        if found then
            //v0 := Solution( Transpose(M), W );
            //V0 := NullspaceOfTranspose(M);
            Wset := { v0 + v : v in V0 };
            Vset := Vset diff Wset;
        end if;
    end for;
    // convert a solution to an element of K
    v := Rep(Vset);
    vprint DQF_verbosity, 2 :  "v =", v;
    alpha := ProductRepresentation(
            [ B[i]^(ZZ ! v[i]) : i in [1..#B] ], 
            [1 : i in [1..#B]]
            );
    vprint DQF_verbosity, 2 :  "alpha =", alpha;

    return alpha;
end function;


function RFF_anisotropic_part_dim2( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 2, 
// this function constructs ist anisotropic part <a,b>.
// This is an implementation of [KR202?, Algorithm 8]

    K := BaseField(q);
    // find the primes that matter
    S  := primes_that_matter(q);  
    
    // make sure the Witt index is even
    if IsOdd(WittIndex(q)) then
        q := q + HyperbolicForm(K, 2);
        q`primes := S;
    end if;
    
    d := Discriminant(q);
    //d := Numerator(d)*Denominator(d);
    
    found := false;
    // we are in u.f.d, so B_ is a set of generators for places in P
    B_ := [ K!Minimum(p) : p in S ];
    // initialization of places' search
    places := PlaceEnumInit(K);
    while not found do
        vprint DQF_verbosity, 2 : "basis:", B_;

        // vector of exponents of Hasse invariants
        W_ := [ZZ | (1 - HasseMinkowskiInvariant(q, P))/2 : P in S ];
        W := Vector(FF2, W_);
        vprint DQF_verbosity, 2 : "W =", W;
    
        // matrix of Hilbert symbols (already transposed => #S cols x #B_ rows)
        H := Matrix(FF2, [[ZZ | (1 - HilbertSymbol(K!b, K!d, P))/2  : P in S ] : b in B_ ]);
        vprint DQF_verbosity, 2 : "H =", H;
        
        // V solution to H*V == W
        found, V, _ := IsConsistent(H, W);
        if found then 
            vprint DQF_verbosity, 2 : "V =", V;
            // the sought element
            alpha := &*[ B_[j]^V_[j] : j in [1..#B_] ] where V_ := Eltseq( ChangeRing(V,ZZ));
        else
            P := PlaceEnumNext(places);
            while P in S do
                P := PlaceEnumNext(places);
            end while;
            vprint DQF_verbosity : "Appending ", P;
            Append(~S, P);
            Append(~B_, K!Minimum(P));
        end if;
    end while;
    vprint DQF_verbosity, 2 : "alpha =", alpha;
    
    //return DiagonalQuadraticForm(K, [alpha, -alpha*d]);
    a := simplify(alpha);
    b := simplify(-alpha*d);
    return DiagonalQuadraticForm(K, [a, b]);
end function;


function anisotropic_part_RFF( q )
// Given a diagonal quadratic form q over a FF_q[x], this function computes its 
//anisotropic part q_a, that is an anisotropic form s.t. q = q_a ⊥ w*<1,-1>.

    K := BaseField(q);

    // simplify the form, replacing every element by its radical
    q := DiagonalQuadraticForm(K, L) 
        where L is [ simplify(c) : c in Coefficients(q) ];

    P := primes_that_matter(q);
    d := AnisotropicDimension(q);
    
    // first short-cut
    if d eq Dimension(q) then
        return q;
    end if;
    
    // initialize the anisotropic part to be an empty form
    qa := DiagonalQuadraticForm(K, []);
    
    // second short-cut
    C := Coefficients(q);
    for c in C do
        q_ := q - DiagonalQuadraticForm(K, [c]);
        // substracting an existing coeff will not change primes
        // don't waste time on a new factorization
        q_`primes := P;
        if AnisotropicDimension(q_) lt d then
            q := q_;
            q`primes := P;
            qa +:= DiagonalQuadraticForm(K, [c]);
            d -:= 1;
        end if;
    end for;
    
    if d eq 4 then
        a_  := DiagonalQuadraticForm(K, [ One(K) ]);
        qa +:= a_;
        q  -:= a_;
        d   := 3;
        assert2 AnisotropicDimension(q) eq 3;
    end if;
  
    if d eq 3 then
        a   := RFF_anisotropic_part_dim3( q );
        a_  := DiagonalQuadraticForm(K, [a]);
        qa +:= a_;
        q  -:= a_;
        d   := 2;
        assert2 AnisotropicDimension(q) eq 2;
    end if;

    if d eq 2 then
        q_ := RFF_anisotropic_part_dim2( q );
        qa +:= q_;
    end if;
    
    if d eq 1 then
        qa +:= DiagonalQuadraticForm(K, [ simplify(disc_q) ])
            where disc_q is Discriminant(q);
    end if;
    
    return qa;
end function;
