freeze;

//////////////////////////////////////////////////
// LegendreSymbol

intrinsic LegendreSymbol( a::FldFunRatElt, P::PlcFunElt ) -> RngIntElt
{ 
Given an element a and a place P of a rational function field, 
this function computes the Legendre symbol of a square-class of a over P. 
}
    K := FunctionField(P);
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";
    require IsCoercible(K, a) : "Incompatible fields!";
    require IsFinite(ConstantField(K)): "The base field must be finite!";

    v := Valuation(a, P);
    if IsOdd(v) then
        return 0;
    end if;
    if v ne 0 then
        p := LocalUniformizer(P);
        a := a*p^(-v);
    end if;
    if IsSquare(Evaluate(a,P)) then
        return 1;
    end if;
    return -1;
end intrinsic;

intrinsic LegendreSymbol( a::FldFunRatElt, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return LegendreSymbol(a, Place(P));
end intrinsic;

//////////////////////////////////////////////////
// IsLocalSquare

intrinsic IsLocalSquare( a::FldFunRatElt, P::PlcFunElt ) -> BoolElt
{
Given an element a in K and a place P of a function field, 
checks whether a is a square at the completion of K at P.
}
    return LegendreSymbol(a,P) eq 1;
end intrinsic;

//////////////////////////////////////////////////
// IsSumOfSquares
// (just for completeness)
intrinsic IsSumOfSquares( a::FldFunRatElt ) -> BoolElt
{ " }
    K := Parent(a);
    require ISA(Category(ConstantField(K)), FldFin) : "Field of constants must be finite!";
    return true;
end intrinsic;

intrinsic IsSOS( a::FldFunRatElt ) -> BoolElt
{ " }
    return IsSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

function length_nondyadic( a, P )
// Given a nonzero element a of a global function field K and a place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  
    
    // the residue field
    Kp := ResidueClassField(P);
    // and valuation
    v := Valuation(a,P);
    // two cases to consider
    if IsOdd(v) then
        // possible values are 2 and 3
        q := DiagonalQuadraticForm(Kp, [Kp!1, Kp!1]);
        return IsIsotropic(q) select 2 else 3;
    else
        // possible values are 1 and 2
        return IsLocalSquare(a, P) select 1 else 2;
    end if;
end function;

intrinsic LengthOfSOS( a::FldFunRatElt ) -> RngIntElt
{
Given a nonzero element a of a global function field K, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    require not IsZero(a): "The argument must be nonzero!";
    // start with trivial cases
    if not IsSOS(a) then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // the possible values now are 2, 3, 4
    K := Parent(a);
    // find all the primes of K that matter for a
    primes := Support(Divisor(a));
    vprint DQF_verbosity : "Primes to be tested:", primes;
    if IsEmpty(primes) then
        // a is a constant
        return LengthOfSOS( ConstantField(K)!a );
    end if;

    n := Max([ length_nondyadic(a, P) : P in primes ]);
    return n;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldFunRatElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;

//////////////////////////////////////////////////
// HilbertSymbol 
// (partially based on Magma code for number fields)

intrinsic HilbertSymbol( a::FldFunRatElt, b::FldFunRatElt, P::PlcFunElt ) -> RngIntElt
{
Given a,b in a global function field and a prime p of K, 
returns the Hilbert symbol (a,b)_P.
}
    require Parent(a) eq Parent(b) : "Arguments must belong to the same field!";
    K := FunctionField(P); 
    require IsCoercible(K, a) and IsCoercible(K, b) : "Incompatible fields!";
    require not IsZero(a) and not IsZero(b): "Arguments must be non zero!";
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";

    va := Valuation(a, P);
    vb := Valuation(b, P);
    p := LocalUniformizer(P);
    a_ := a*p^(-va);
    b_ := b*p^(-vb);
    
    // four cases two consider
    if IsEven(va) and IsEven(vb) then 
        return 1;
    elif IsEven(va) and IsOdd(vb) then
        return IsSquare(Evaluate(a_,P)) select 1 else -1;
    elif IsOdd(va) and IsEven(vb) then
        return IsSquare(Evaluate(b_,P)) select 1 else -1;
    else
        return IsSquare(Evaluate(-a_*b_,P)) select 1 else -1;
    end if;
end intrinsic;

intrinsic HilbertSymbol( a::FldFunRatElt, b::FldFunRatElt, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HilbertSymbol(a, b, Place(P));
end intrinsic

intrinsic HilbertSymbol( a::RngUPolElt, b::RngUPolElt, P::PlcFunElt ) -> RngIntElt
{ " } //"
    K := FunctionField(P);
    return HilbertSymbol( K!a, K!b, P );
end intrinsic

intrinsic HilbertSymbol( a::RngUPolElt, b::RngUPolElt, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HilbertSymbol(a, b, Place(P));
end intrinsic

