freeze;

import "../helpers/varia.m" : ZZ, QQ, FF2;
import "quadratic_forms.m" : primes_that_matter;

function QQ_anisotropic_part_dim4( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension d ≥ 4, 
// this function constructs an element α ∈ QQ such that dim_a( q ⊥ <−α>) = d − 1.

    d := AnisotropicDimension(q);
    sign_q := &+ [ Sign(c) : c in Coefficients(q) ];
    if sign_q gt 0 then
        alpha := 1;
    else 
        alpha := -1;
    end if;
    return alpha;
end function;

function QQ_anisotropic_part_dim3( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 3, 
// this function constructs an element α ∈ QQ, such that dim_a( q ⊥ <−α>) = 2.
// This is an implementation of [KR202?, Algorithm 7]

    // find the primes that matter
    S  := primes_that_matter(q);
    d  := Discriminant(q);
    
    // alpha must be congruent to either d-1 (mod p) or p (mod p^2)
    S1 := [ p : p in S | IsOdd(Valuation(d, p)) ];
    E1 := [ ZZ!d-1 : p in S1 ];
    S2 := [ p^2 : p in S | IsEven(Valuation(d, p)) ];
    E2 := [ p : p in S2 ];
    alpha := ChineseRemainderTheorem(E1 cat E2, S1 cat S2);
    
    // if the sign matters, then the sign of alpha must match the sign of q
    sign_q := &+ [ Sign(c) : c in Coefficients(q) ];
    if sign_q eq -3 then
        alpha -:= Lcm(S1 cat S2);
    end if;
    vprint DQF_verbosity : "alpha =", alpha;
    
    a, _ := SquareFreeFactorization(alpha);
    return a;
end function;


function QQ_anisotropic_part_dim2( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 2, 
// this function constructs ist anisotropic part <a,b>.
// This is a version for real fields
// This is an implementation of [KR202?, Algorithm 8]

    // find the primes that matter
    S  := primes_that_matter(q);   
    
    // make sure the Witt index is divisible by 4
    w := (WittIndex(q) mod 4);
    if w ne 0 then
        q := q + HyperbolicForm(QQ, 2*(4-w));
        q`primes := S;
    end if;
    d := Discriminant(q);
    d := Numerator(d)*Denominator(d);
    
    found := false;
    // initialization of places' search
    p := 2;
    while not found do
        // na czuja biorę B = S
        B_ := S;
        vprint DQF_verbosity : "basis:", B_;

        // does sign matters?
        if d lt 0 then
            W_ := [ZZ | 1/2 - s/4] where s is &+ [ Sign(c) : c in Coefficients(q) ];
        else 
            W_ := [ZZ | ];
        end if;
        // vector of exponents of Hasse invariants
        W_ := W_ cat [ZZ | (1 - HasseMinkowskiInvariant(q, P))/2 : P in S ];
        W := Vector(FF2, W_);
        vprint DQF_verbosity : "W =", W;
    
        // matrix of Hilbert symbols (already transposed => #S cols x #B_ rows)
        B := Matrix(FF2, [[ZZ | (1 - HilbertSymbol(b, d, P))/2  : P in S ] : b in B_ ]);
        
        // does sign matters?
        if d lt 0 then
            A := Matrix(FF2, [[ Sign(P) eq 1 select 0 else 1 ] : P in S ]);
            // join the two matrices
            AB := HorizontalJoin(A, B);
        else
            AB := B;
        end if;
        vprint DQF_verbosity : "AB =", AB;
        
        // V solution to H*V == W
        found, V, _ := IsConsistent(AB, W);
        vprint DQF_verbosity : "V =", V;
        
        if found then
            // the sought element
            alpha := &*[ B_[j]^V_[j] : j in [1..#B_] ] where V_ := Eltseq( ChangeRing(V,ZZ));
        else
            p := NextPrime(p);
            while p in S do
                p := NextPrime(p);
            end while;
            Append(~S, p);
        end if;
    end while;
    
    a, _ := SquareFreeFactorization(alpha);
    b, _ := SquareFreeFactorization(-alpha*d);
    return DiagonalQuadraticForm(QQ, [a, b]);
end function;

function anisotropic_part_QQ( q )
// Given a diagonal quadratic form q over the rationals, this function computes its 
//anisotropic part q_a, that is an anisotropic form s.t. q = q_a ⊥ w*<1,-1>.

    // simplify the form, replacing every element by its radical
    q := DiagonalQuadraticForm(QQ, L) 
        where L is [ SquareFreeFactorization(Numerator(a)*Denominator(a)) : a in C ] 
        where C is Coefficients(q);

    P := primes_that_matter(q);
    d := AnisotropicDimension(q);
    
    // first short-cut
    if d eq Dimension(q) then
        return q;
    end if;
    
    // initialize the anisotropic part to be an empty form
    qa := DiagonalQuadraticForm(QQ, []);
    
    // second short-cut
    C := Coefficients(q);
    for c in C do
        q_ := q - DiagonalQuadraticForm(QQ, [c]);
        // substracting an existing coeff will not change primes
        // don't waste time on a new factorization
        q_`primes := P;
        if AnisotropicDimension(q_) lt d then
            q := q_;
            q`primes := P;
            qa +:= DiagonalQuadraticForm(QQ, [c]);
            d -:= 1;
        end if;
    end for;
    
    while d ge 4 do
        a   := QQ_anisotropic_part_dim4( q );
        a_  := DiagonalQuadraticForm(QQ, [a]);
        qa +:= a_;
        q  -:= a_;
        d  -:= 1;
    end while;
    
    if d eq 3 then
        a   := QQ_anisotropic_part_dim3( q );
        a_  := DiagonalQuadraticForm(QQ, [a]);
        qa +:= a_;
        q  -:= a_;
        d   := 2;
    end if;

    if d eq 2 then
        q_ := QQ_anisotropic_part_dim2( q );
        qa +:= q_;
    end if;
    
    if d eq 1 then
        disc_q := Discriminant(q);
        disc_q := Numerator(disc_q)*Denominator(disc_q);
        a, _ := SquareFreeFactorization(disc_q);
        qa +:= DiagonalQuadraticForm(QQ, [a]);
    end if;
    
    return qa;
end function;
