freeze;

declare attributes DiagQuadFrm: primes;

import "../number_fields/quadratic_forms.m" : is_locally_isotropic_nondyadic;
import "../helpers/varia.m" : ZZ, QQ;


//////////////////////////////////////////////////
// primes_that_matter

function primes_that_matter( q )
// find all the primes of K that matter for q

    if not assigned q`primes then
        q`primes := &join[ Seqset(PrimeDivisors(Numerator(c)*Denominator(c))) : c in q`coeffs ];
        q`primes := [2] cat [ p : p in q`primes ];
    end if;
    return q`primes;
end function;


//////////////////////////////////////////////////
// HasseMinkowskiInvariant
// WittInvariant

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::RngIntElt ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";

    d := Dimension(q);
    L := [ HilbertSymbol(q`coeffs[i],q`coeffs[j], P) : j in [(i+1)..d] , i in [1..d-1] ];
    return &* L;
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::RngIntElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, P);
end intrinsic;


//////////////////////////////////////////////////
// IsIsotropic

function is_locally_isotropic_QQ2( q )
    // Given a quadratic form q over QQ 
    // this function checks if the localization of q at 2 is isotropic.

    // the dimension
    d := Dimension(q);
    // the determinant
    detq := &* q`coeffs;
    // only square class matters
    detq := Numerator(detq)*Denominator(detq);
    //QQ2 := pAdicField(2);
    
    if d le 1 then
        return false;
    elif d ge 5 then
        return true;
    elif d eq 2 then
        return IsLocalSquare(-detq, 2);
    elif d eq 3 then
        // use [Lam, Prop. V.3.22]
        s := HilbertSymbol(-1, -detq, 2);
        sq := WittInvariant(q, 2);
        return (s eq sq);
    elif IsLocalSquare(detq, 2) then
        // if we get here, then d == 4        
        // for quartic form with determinant 1, use [ Lam, Proposition V.3.23]
        s := HilbertSymbol(-1, -1, 2);
        sq := WittInvariant(q, 2);
        vprint DQF_verbosity: "s =", s, "sq =", sq;
        return (s eq sq);
    end if;
    // form is unique with det a square
    return true;
end function;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngInt ) -> BoolElt
{
Checks if a quadratic form q over QQ is locally isotropic at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    if (QQ!2) notin P then
        return is_locally_isotropic_nondyadic(q,P);
    else
        return is_locally_isotropic_QQ2(q,P);
    end if;
end intrinsic;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngIntElt ) -> BoolElt
{
Checks if a quadratic form q over QQ is locally isotropic at P.
}
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    if P ne 2 then
        return is_locally_isotropic_nondyadic(q,P*ZZ);
    else
        return is_locally_isotropic_QQ2(q);
    end if;
end intrinsic;

function is_isotropic_QQ( q )
    // Given a quadratic form q with coefficients in QQ,
    // this function tests if q is isotropic.

    // the base field and its maximal order
    assert q`base_ring eq QQ;

    // the dimension
    d := Dimension(q);
    if d le 1 then
        return false;
    elif d eq 2 then
        return IsSquare(Discriminant(q));
    end if;
    vprint DQF_verbosity : "General dimension.";

    // compute the signature of q
    if forall{ c : c in q`coeffs | c gt 0} or forall{ c : c in q`coeffs | c lt 0} then
        return false;
    end if;
    vprint DQF_verbosity : "Signature test passed.";
          
    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         q`primes := &join[ Seqset(PrimeDivisors(Numerator(c)*Denominator(c))) : c in q`coeffs ];
//         q`primes := [2] cat [ P : P in q`primes ];
//     end if;
    PP := primes_that_matter(q);
    vprint DQF_verbosity : "Primes to be tested:", PP;
    // check that q is locally isotropic over all of them
    vprint DQF_verbosity : "Local results:", [ <p, IsIsotropic(q, p)> : p in PP ];
    return forall{ p : p in PP | IsIsotropic(q, p) }; 
end function;


//////////////////////////////////////////////////
// IsHyperbolic

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngIntElt ) -> BoolElt
{
Checks if a quadratic form q over QQ is locally hyperbolic at P.
}
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsLocalSquare(Discriminant(q), P) then
        return false;
    end if;
    m := ZZ ! (d/2);
    K := q`base_ring;
    t := HilbertSymbol(-1,-1,P)^( ZZ ! (m^2-m)/2 );
    return HasseMinkowskiInvariant(q, P) eq t;
end intrinsic;

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngInt ) -> BoolElt
{ " } //"
    return IsHyperbolic(q, Generator(P));
end intrinsic;

function is_hyperbolic_QQ( q )
    // Given a quadratic form q over rationals, 
    // this function returns true if q is hyperbolic and false otherwise.
    
    // the base field and its maximal order
    //K := q`base_ring;
    //OK := MaximalOrder(K);

    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsSquare(Discriminant(q)) then
        return false;
    end if;

    // compute the signature of q
    if not IsZero( &+ [ Sign(c) : c in q`coeffs ] ) then
        return false;
    end if;
    vprint DQF_verbosity : "Signature test passed.";
          
    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         q`primes := &join[ Seqset(PrimeDivisors(Numerator(c)*Denominator(c))) : c in q`coeffs ];
//         q`primes := [2] cat [ P : P in q`primes ];
//     end if;
    PP := primes_that_matter(q);
    vprint DQF_verbosity : "Primes to be tested:", PP;
    // check that q is locally hyperbolic over all of them
    vprint DQF_verbosity : "Local results:", [ <p, IsHyperbolic(q, p)> : p in PP ];
    return forall{ p : p in PP | IsHyperbolic(q, p) };
end function;

//////////////////////////////////////////////////
// AnisotropicDimension

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngInt ) -> RngIntElt
{
Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. The form q_1 is called an anisotropic part of q.

Given a quadratic form q and a prime P, this function returns a dimension 
of an anisotropic part of localilization of q at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    
    d := Dimension(q);
    if IsEven(d) then
        // dimension of hyperbolic form is 0
        if IsHyperbolic(q, P) then
            return 0;
        elif IsLocalSquare(Discriminant(q), P) then
            return 4;
        else 
            return 2;
        end if;
    end if;
    // when q is odd-fimensional, extend it to the next even dimension
    c := QQ ! (-1)^(ZZ ! (d*(d+1)/2));
    q2 := DiagonalQuadraticForm( q`base_ring, q`coeffs cat [ c * &* q`coeffs ] );
    if IsHyperbolic( q2, P ) then
        return 1;
    else
        return 3;
    end if;
end intrinsic;

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngIntElt ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    return AnisotropicDimension(q, P*ZZ);
end intrinsic;

function anisotropic_dimension_QQ( q )
// Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
// as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. 
// The form q_1 is called an anisotropic part of q.
//
// Given a quadratic form q over a number field K, 
// this function returns a dimension of an anisotropic part of q.

    // compute the signatures of q in all real embedding of K
    N := Abs( &+ [ Sign(c) : c in q`coeffs ] );
    vprint DQF_verbosity : "Signature: ", N;
    // for n > 2, it must be a real place, where the the anisotropic
    // part has a maximal dimension
    if N  ge 3 then
        return N;
    end if;
    vprint DQF_verbosity : "Signatures passed";
          
    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         q`primes := &join[ Seqset(PrimeDivisors(Numerator(c)*Denominator(c))) : c in q`coeffs ];
//         q`primes := [2] cat [ P : P in q`primes ];
//     end if;
    PP := primes_that_matter(q);
    vprint DQF_verbosity : "Primes to be tested:", PP;
    // compute maximal anisotropic dimension over all primes
    M :=  Max([ AnisotropicDimension(q, p) : p in PP ]);
    return Max(M, N); 
end function;


//////////////////////////////////////////////////
// IsTorsion

function is_torsion_QQ( _q )
// Check if the element _q of a Witt ring is torsion.
    q := _q`form;
    sgn_q := Abs( &+ [ Sign(c) : c in q`coeffs ] );
    return IsZero(sgn_q);
end function;

